-------------------------------------------------------------------------------
--
-- VGA raster display for Imagic
-- This also acts as an SRAM controller
-- Since the VGA must read from the SRAM at a rate
-- of 25 MHZ, the VGA controller is given full
-- control of the SRAM.  If this nios wants to write
-- data to the SRAM, it must ask the VGA controller to do so.
--
--
--  From code produced by Stephen A. Edwards, 
--  Columbia University, sedwards@cs.columbia.edu
--
-------------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

entity vga_raster is
  
  port (
-- ++ WED FINAL PROJECT
    -- Add Bus Interface
    avs_s1_clk        : in  std_logic;
    avs_s1_reset_n    : in  std_logic;
    avs_s1_read       : in  std_logic;
    avs_s1_write      : in  std_logic;
    avs_s1_chipselect : in  std_logic;
    avs_s1_waitrequest      : out  std_logic;
    avs_s1_address    : in  std_logic_vector(4 downto 0);
    avs_s1_readdata   : out std_logic_vector(15 downto 0);
    avs_s1_writedata  : in  std_logic_vector(15 downto 0);

    -- Control Singals for SRAM
    SRAM_DQ   : inout std_logic_vector(15 downto 0);
    SRAM_ADDR : out std_logic_vector(17 downto 0);
    SRAM_UB_N,
    SRAM_LB_N,
    SRAM_WE_N,
    SRAM_CE_N,
    SRAM_OE_N : out std_logic;    
-- -- WED FINAL PROJECT

    reset : in std_logic;

    VGA_CLK,                                 -- Clock
    VGA_HS,                                  -- H_SYNC
    VGA_VS,                                  -- V_SYNC
    VGA_BLANK,                               -- BLANK
    VGA_SYNC : out std_logic;                -- SYNC
    VGA_R,                                   -- Red[9:0]
    VGA_G,                                   -- Green[9:0]
    VGA_B : out std_logic_vector(9 downto 0) -- Blue[9:0]
  );

end vga_raster;

architecture rtl of vga_raster is
  -- Video parameters
  
  constant HTOTAL       : integer := 800;
  constant HSYNC        : integer := 96;
  constant HBACK_PORCH  : integer := 48;
  constant HACTIVE      : integer := 640;
  constant HFRONT_PORCH : integer := 16;
  
  constant VTOTAL       : integer := 525;
  constant VSYNC        : integer := 2;
  constant VBACK_PORCH  : integer := 33;
  constant VACTIVE      : integer := 480;
  constant VFRONT_PORCH : integer := 10;
  constant MID_X		: integer := 320;
  constant MID_Y        : integer := 240;

  -- Signals for the video controller
  signal Hcount : std_logic_vector(9 downto 0);  -- Horizontal position (0-800)
  signal Vcount : std_logic_vector(9 downto 0);  -- Vertical position (0-524)
  signal EndOfLine, EndOfField : std_logic;

-- ++ WED FINAL_PROJECT
  signal IMAGE_HEND      : std_logic_vector(8 downto 0) := "000000001";   -- Image Width from NIOS
  signal IMAGE_VEND      : std_logic_vector(8 downto 0) := "000000001";   -- Image Height from NIOS
  signal IMAGE_RED       : std_logic_vector(4 downto 0) := "00000";       -- Red Pixel data from NIOS
  signal IMAGE_GREEN     : std_logic_vector(4 downto 0) := "00000";       -- Green Pixel data from NIOS
  signal IMAGE_BLUE      : std_logic_vector(4 downto 0) := "00000";       -- Blue Pixel data from NIOS
  signal BG_RED          : std_logic_vector(4 downto 0) := "00000";       -- Red background data from NIOS
  signal BG_GREEN        : std_logic_vector(4 downto 0) := "00000";       -- Green background data from NIOS
  signal BG_BLUE         : std_logic_vector(4 downto 0) := "00000";       -- Blue background data from NIOS
  signal IMAGE_HSTART : integer;                  -- Used during display processing
  signal IMAGE_VSTART : integer;                  -- Used during display processing
  signal image_h, image_v, image : std_logic;     -- Image area
  signal clk                           : std_logic := '0';    -- Internally used 25MHz clock
  signal RAM_ADDR                      : std_logic_vector(17 downto 0) := "000000000000000000";
  signal RAM_WRITE_ADDR                : std_logic_vector(17 downto 0) := "000000000000000000";
  signal write_data, write_background  : std_logic := '0';  -- Flag
  -- BUFFERS for SRAM
  signal ADDRESS_TO_SRAM               : std_logic_vector(17 downto 0) := "000000000000000000";
  signal WRITEDATA_TO_SRAM             : std_logic_vector(15 downto 0) := "0000000000000000";
  signal READDATA_FROM_SRAM            : std_logic_vector(15 downto 0) := "0000000000000000";  
  signal WRITEENABLE_TO_SRAM           : std_logic := '0';
-- -- WED FINAL_PROJECT

  signal vga_hblank, vga_hsync,
  	 vga_vblank, vga_vsync : std_logic;  -- Sync. signals

begin

-- ++ WED FINAL_PROJECT
  -- If writing data, force SRAM bus, otherwise let it float
  SRAM_DQ <= WRITEDATA_TO_SRAM when WRITEENABLE_TO_SRAM = '0' else
             (others => 'Z');
  -- If data is on SRAM bus, store it
  READDATA_FROM_SRAM <= SRAM_DQ;
  -- Address going out to SRAM
  SRAM_ADDR <= ADDRESS_TO_SRAM;
  -- Write Enable Signal going out to SRAM
  SRAM_WE_N <= WRITEENABLE_TO_SRAM;
  -- These can all be set to zero.  Only need simple mode of operation.  
  SRAM_UB_N <= '0';
  SRAM_LB_N <= '0';
  SRAM_CE_N <= '0';
  SRAM_OE_N <= '0';
  
  -- Split the clock for the VGA (from 50 Mhz to 25 Mhz)
  process (avs_s1_clk)
  begin
    if avs_s1_clk'event and avs_s1_clk = '1' then
      clk <= not clk;
    end if;
  end process;

  -- State Machine handling SRAM control and also accepting commands from
  -- NIOS
  -- There are three 'periods' of operation
  --   Period 1)  Change SRAM read address
  --              Raise any flags set during write period.  The low to high
  --              transition is what actually writes the data to the SRAM
  --   Period 2)  Accept commands from NIOS.
  --              This includes accepting data to be written to SRAM, accepting
  --              image height/width that is written to buffers, and SRAM
  --              write address that is also stored in a buffer.
  --              The actual writing of SRAM is initiated here as well.  
  --   Period 3)  Read pixel data from SRAM.  If the SRAM has not recently be written
  --              read data from SRAM to be displayed on screen.  If the SRAM has
  --              been written, there is not enough time for the SRAM to drive the bus
  --              so we just display the previous pixel again (the human eye can't tell)
  SRAM_CONTROLLER : process(avs_s1_clk, avs_s1_reset_n, reset)
  begin
    if reset = '1' then
      avs_s1_readdata <= (others => '0'); 
    elsif avs_s1_clk'event and avs_s1_clk = '1' then    
       if clk = '1' then   --(period 1)
         WRITEENABLE_TO_SRAM <= '1';   -- This is the READ portion, so disable writes 
         avs_s1_waitrequest <= '0';    -- Tell bus during next cycle data can be processed     
         write_data <= '0';            -- Always reset flag, this is READ portion
         ADDRESS_TO_SRAM <= RAM_ADDR;  -- Update the pixel that will be read during down tick
       else                --(period 2)
         -- check if data should be written
         if avs_s1_write = '1' then               -- They want to write something
           if avs_s1_address = "00000" then       -- Data
             ADDRESS_TO_SRAM <= RAM_WRITE_ADDR;   -- Send SRAM address we had stored in buffer
             write_data <= '1';                   -- Raise flag, DONT read anything, we have to drive SRAM bus
             WRITEDATA_TO_SRAM <= avs_s1_writedata(15 downto 0);     --What SRAM Data?
             WRITEENABLE_TO_SRAM <= '0';          -- Tell SRAM to start writing, we can do this right away
             if write_background = '1' then
                 write_background <= '0';
             -- If this is the first pixel of the image, save it for the background
                 BG_RED <= avs_s1_writedata(14 downto 10);
                 BG_GREEN <= avs_s1_writedata(9 downto 5);
                 BG_BLUE <= avs_s1_writedata(4 downto 0);
             end if;
           elsif avs_s1_address = "00001" then    -- Address (lower 16 bits)
             RAM_WRITE_ADDR(15 downto 0) <= avs_s1_writedata(15 downto 0); -- store SRAM address they want to write to in buffer
		   elsif avs_s1_address = "00010" then    -- Address (upper 2 bits)        
             RAM_WRITE_ADDR(17 downto 16) <= avs_s1_writedata(1 downto 0);
           elsif avs_s1_address = "00011" then    -- Image Width    
             IMAGE_HEND <= avs_s1_writedata(8 downto 0);
             -- If the image size is '1', assume the image is being written to SRAM
             -- and next data point will be the first pixel point, and that will be used as the backgroud.
             if avs_s1_writedata(8 downto 0) = "000000001" then
                write_background <= '1';
             end if;
           elsif avs_s1_address = "00100" then    -- Image Height
             IMAGE_VEND <= avs_s1_writedata(8 downto 0);
           end if;
         end if;
       avs_s1_waitrequest <= '1';   -- Tell bus during next cycle data can NOT be processed, they have to wait          
       end if;
    elsif avs_s1_clk'event and avs_s1_clk = '0' then
      if clk = '1' and write_data = '0' then   -- Read Data from SRAM (period 3)
        IMAGE_RED <= READDATA_FROM_SRAM(14 downto 10);
        IMAGE_GREEN <= READDATA_FROM_SRAM(9 downto 5);
        IMAGE_BLUE <= READDATA_FROM_SRAM(4 downto 0);
      end if;
    end if;
  end process SRAM_CONTROLLER;
  -- Convert data read from NIOS into a useful form for us to use.   
  IMAGE_HSTART <= MID_X - conv_integer(IMAGE_HEND(8 downto 1));
  IMAGE_VSTART <= MID_Y - conv_integer(IMAGE_VEND(8 downto 1)); 
-- -- WED FINAL_PROJECT

  HCounter : process (clk, reset)
  begin    
    if reset = '1' then
      Hcount <= (others => '0');
    elsif clk'event and clk = '1' then
      if EndOfLine = '1' then
        Hcount <= (others => '0');
      else
        Hcount <= Hcount + 1;
      end if;      
    end if;
  end process HCounter;

  EndOfLine <= '1' when Hcount = HTOTAL - 1 else '0';
 
  VCounter: process (clk, reset)
  begin
    if reset = '1' then
      Vcount <= (others => '0');
    elsif clk'event and clk = '1' then
      if EndOfLine = '1' then
        if EndOfField = '1' then
          Vcount <= (others => '0');
        else
          Vcount <= Vcount + 1;
        end if;
      end if;
    end if;
  end process VCounter;

  EndOfField <= '1' when Vcount = VTOTAL - 1 else '0';

  -- State machines to generate HSYNC, VSYNC, HBLANK, and VBLANK
  HSyncGen : process (clk, reset)
  begin    
    if reset = '1' then
      vga_hsync <= '1';
    elsif clk'event and clk = '1' then
      if EndOfLine = '1' then
        vga_hsync <= '1';
      elsif Hcount = HSYNC - 1 then
        vga_hsync <= '0';
      end if;      
    end if;
  end process HSyncGen;
  
  HBlankGen : process (clk, reset)
  begin    
    if reset = '1' then
      vga_hblank <= '1';
    elsif clk'event and clk = '1' then
      if Hcount = HSYNC + HBACK_PORCH then
        vga_hblank <= '0';
      elsif Hcount = HSYNC + HBACK_PORCH + HACTIVE then
        vga_hblank <= '1';
      end if;      
    end if;
  end process HBlankGen;

  VSyncGen : process (clk, reset)
  begin    
    if reset = '1' then
      vga_vsync <= '1';
    elsif clk'event and clk = '1' then
      if EndOfLine ='1' then
        if EndOfField = '1' then
          vga_vsync <= '1';
        elsif Vcount = VSYNC - 1 then
          vga_vsync <= '0';
        end if;
      end if;      
    end if;
  end process VSyncGen;

  VBlankGen : process (clk, reset)
  begin    
    if reset = '1' then
      vga_vblank <= '1';
    elsif clk'event and clk = '1' then
      if EndOfLine = '1' then
        if Vcount = VSYNC + VBACK_PORCH - 1 then
          vga_vblank <= '0';
        elsif Vcount = VSYNC + VBACK_PORCH + VACTIVE - 1 then
          vga_vblank <= '1';
        end if;
      end if;      
    end if;
  end process VBlankGen;

  -- Control the position of the image

  ImageHGen : process (clk, reset)
  begin    
    if reset = '1' then
      image_h <= '1';
    elsif clk'event and clk = '1' then
-- ++ WED FINAL PROJECT
      if Hcount = HSYNC + HBACK_PORCH + IMAGE_HSTART
      then  
        -- Starting displaying image 
        image_h <= '1';
      elsif Hcount = HSYNC + HBACK_PORCH + IMAGE_HSTART + conv_integer(IMAGE_HEND) then
        -- Stop displaying image   
        image_h <= '0';
-- -- WED FINAL_PROJECT
      end if;      
    end if;
  end process ImageHGen;

  ImageVGen : process (clk, reset)
  begin
    if reset = '1' then
      image_v <= '0';
    elsif clk'event and clk = '1' then
      if EndOfLine = '1' then
-- ++ WED FINAL_PROJECT
        if Vcount = VSYNC + VBACK_PORCH - 1 + IMAGE_VSTART
        then
          -- Starting displaying image
          image_v <= '1';
        elsif Vcount = VSYNC + VBACK_PORCH - 1 + IMAGE_VSTART + conv_integer(IMAGE_VEND) then
          -- Stop displaying image
          image_v <= '0';
-- -- WED FINAL_PROJECT
        end if;
      end if;      
    end if;
  end process ImageVGen;

-- ++ WED FINAL_PROJECT
  -- If the pixel that is being written to the
  -- screen is within image region, set image flag
  image <= image_h and image_v;

  -- Every 25 MHz, if image is being display, increment
  -- the address in the SRAM.  Since the pixels
  -- are being display continously, this gets the next
  -- pixel that needs to be displayed.  (increment by
  -- two since address is byte addressed).  When
  -- the screen reachs the lower right pixel, reset
  -- the ram address.  
  RamAddr : process (clk, reset)
  begin
    if clk'event and clk = '0' then
       if EndOfField = '1' and EndofLine= '1' then
          RAM_ADDR <= "000000000000000000";  -- reset address
       elsif image_h = '1' and image_v = '1' then
          RAM_ADDR <= RAM_ADDR + 2;          -- increment
       end if;
    end if;
  end process RamAddr;
-- -- WED FINAL_PROJECT

  -- Registered video signals going to the video DAC

  VideoOut: process (clk, reset)
  begin
    if reset = '1' then
      VGA_R <= "0000000000";
      VGA_G <= "0000000000";
      VGA_B <= "0000000000";
    elsif clk'event and clk = '1' then
-- ++ WED FINAL_PROJECT
      if image = '1' then   
        if IMAGE_HEND = "000000001" then
            -- if image height is 1, assume background should be shown
            VGA_R(8 downto 4) <= BG_RED;
            VGA_G(8 downto 4) <= BG_GREEN;
            VGA_B(8 downto 4) <= BG_BLUE;
            VGA_R(3 downto 0) <= BG_RED(4 downto 1);
            VGA_G(3 downto 0) <= BG_GREEN(4 downto 1);
            VGA_B(3 downto 0) <= BG_BLUE(4 downto 1); 
        else
            -- Copy Data from image into lower and upper halves of pixel output
            VGA_R(8 downto 4) <= IMAGE_RED;
            VGA_G(8 downto 4) <= IMAGE_GREEN;
            VGA_B(8 downto 4) <= IMAGE_BLUE;
            VGA_R(3 downto 0) <= IMAGE_RED(4 downto 1);
            VGA_G(3 downto 0) <= IMAGE_GREEN(4 downto 1);
            VGA_B(3 downto 0) <= IMAGE_BLUE(4 downto 1);
        end if;
      elsif vga_hblank = '0' and vga_vblank ='0' then
        -- Set background of image to first pixel value of image
        VGA_R(8 downto 4) <= BG_RED;
        VGA_G(8 downto 4) <= BG_GREEN;
        VGA_B(8 downto 4) <= BG_BLUE;
        VGA_R(3 downto 0) <= BG_RED(4 downto 1);
        VGA_G(3 downto 0) <= BG_GREEN(4 downto 1);
        VGA_B(3 downto 0) <= BG_BLUE(4 downto 1);
-- -- WED FINAL_PROJECT
      else
        VGA_R <= "0000000000";
        VGA_G <= "0000000000";
        VGA_B <= "0000000000";    
      end if;
    end if;
  end process VideoOut;

  VGA_CLK <= clk;
  VGA_HS <= not vga_hsync;
  VGA_VS <= not vga_vsync;
  VGA_SYNC <= '0';
  VGA_BLANK <= not (vga_hsync or vga_vsync);

end rtl;
